/*
 Copyright (C)2004 Dimax ( http://www.xdimax.com )

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
*/
  
/*
*-------------------------------------------------------------------
* Description: i2cbrdg main module
*-------------------------------------------------------------------								 
* Version: $Id: i2cbrdg.c,v 1.1 2005/01/21 17:08:22 xol Exp $									   
*-------------------------------------------------------------------
*/
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include <usb.h>
#include "i2cbrdg.h"


//-----------------------------------------------------------------------------
// Types 
//-----------------------------------------------------------------------------

#define MAX_INPUT_BUFFER_LENGTH     257 // Maximum length of the data we are waiting from device
#define MAX_OUTPUT_BUFFER_LENGTH    264 // Maximum length of the data we are sending to device

typedef struct _DEVICE_REQUEST
{
    UCHAR InputBuffer[MAX_INPUT_BUFFER_LENGTH]; // Buffer received from device
    ULONG InputBufferLength;    // Length of the buffer received from device
    UCHAR OutputBuffer[MAX_OUTPUT_BUFFER_LENGTH]; // Buffer sent to device
    ULONG OutputBufferLength;   // Length of the buffer sent to device
} DEVICE_REQUEST, *PDEVICE_REQUEST;


//-----------------------------------------------------------------------------
// Constants
//-----------------------------------------------------------------------------

#define DIMAX_VENDOR    0x0abf
#define DIMAX_PRODUCT   0x3370
#define DIMAX_IF        0x00
#define DIMAX_IN_EP     0x84
#define DIMAX_OUT_EP    0x02

#define DIMAX_TIMEOUT   100     // ms

int debug=0;

//-----------------------------------------------------------------------------
// Global Variables
//-----------------------------------------------------------------------------
static int usb_init_done;

//-----------------------------------------------------------------------------
// Local Function Definition
//-----------------------------------------------------------------------------
static U2C_RESULT u2c_libusb_rc( int rc );
static U2C_RESULT u2c_fw_rc( int rc );
static int u2c_transaction( HANDLE hDevice, PDEVICE_REQUEST req );


//-----------------------------------------------------------------------------
// Open/Close Functions 
//-----------------------------------------------------------------------------
BYTE U2C_GetDeviceCount( void )
{
    BYTE    dev_count=0;
    struct  usb_bus *bus;
    struct  usb_device *dev;

    if( ! usb_init_done )
    {
        usb_init();
        usb_init_done=1;
    }

    usb_find_busses();
    usb_find_devices();
    bus=usb_get_busses();

    for( ; bus; bus=bus->next )
        for( dev=bus->devices; dev; dev=dev->next )
        {
            if ( dev->descriptor.idVendor  == DIMAX_VENDOR &&
                 dev->descriptor.idProduct == DIMAX_PRODUCT ) 
            {
                dev_count++;
            }
        }
        
    return dev_count;
}


HANDLE U2C_OpenDevice(BYTE nDevice)
{
    BYTE    devc=0;
    int     found=0;
    struct  usb_bus *bus;
    struct  usb_device *dev=0;
    HANDLE  hDevice;

    if( ! usb_init_done )
    {
        usb_init();
        usb_init_done=1;
    }

    usb_find_busses();
    usb_find_devices();
    bus=usb_get_busses();

    while(bus && !found)
    {
        for( dev=bus->devices; dev; dev=dev->next )
        {
            if ( dev->descriptor.idVendor  == DIMAX_VENDOR &&
                 dev->descriptor.idProduct == DIMAX_PRODUCT ) 
            {
                if( devc == nDevice )
		        {
        		    found = 1;
                    break;
        		}
                devc++;
            }
        }
        bus = bus->next;
    }
    
    if( ! dev )
        return INVALID_HANDLE_VALUE;

    /* 
    * Open device file 
    * You should have sufficient permisions to do that
    */
    hDevice = usb_open(dev);
    if( ! hDevice )
        return INVALID_HANDLE_VALUE;

    return hDevice;
}


U2C_RESULT U2C_CloseDevice(HANDLE hDevice)
{
    usb_close( hDevice );
    return U2C_SUCCESS;
}


U2C_RESULT U2C_IsHandleValid(HANDLE hDevice)
{
    if( hDevice )
        return U2C_SUCCESS;
    else
        return U2C_HARDWARE_NOT_FOUND;
}    


U2C_RESULT U2C_GetSerialNum(HANDLE hDevice, long* pSerialNum)
{
    U2C_RESULT rc; 
    DEVICE_REQUEST req;

    req.OutputBufferLength=1;
    req.OutputBuffer[0]=0x0B;
    req.InputBufferLength=5;

    rc = u2c_transaction( hDevice, &req );
    memcpy(pSerialNum, req.InputBuffer, 4);
    return rc;
}


U2C_RESULT U2C_GetFirmwareVersion(HANDLE hDevice, PU2C_VERSION_INFO pVersion)
{
    U2C_RESULT rc; 
    DEVICE_REQUEST req;

    req.OutputBufferLength=1;
    req.OutputBuffer[0]=0x0A;
    req.InputBufferLength=3;

    rc = u2c_transaction( hDevice, &req );
    pVersion->MajorVersion = req.InputBuffer[0];
    pVersion->MinorVersion = req.InputBuffer[1];
    return rc;
}


//-----------------------------------------------------------------------------
// I2C High Level Functions 
//-----------------------------------------------------------------------------
U2C_RESULT U2C_Read(HANDLE hDevice, PU2C_TRANSACTION pTransaction)
{
    U2C_RESULT rc;
    DEVICE_REQUEST req;
    long    MemoryAddr;
    int     j,i = 0;
    
    if (pTransaction->nSlaveDeviceAddress > 0x7F)
        return U2C_BAD_PARAMETER;

    req.OutputBuffer[i++] = 0x01;
    req.OutputBuffer[i++] = pTransaction->nMemoryAddressLength;
    MemoryAddr = pTransaction->nMemoryAddress;
    
    if (pTransaction->nMemoryAddressLength != 0)
    {
        // slave address for write (lsb low)
        req.OutputBuffer[i++] = (pTransaction->nSlaveDeviceAddress<<1) & 0xFE;
        for ( j = i + pTransaction->nMemoryAddressLength - 1; j >= i; j--)
        {
            req.OutputBuffer[j] = (BYTE)MemoryAddr & 0xFF;
            MemoryAddr >>= 8;
        }
    }
    i += pTransaction->nMemoryAddressLength;
    
    // slave address for read (with lsb hi)
    req.OutputBuffer[i++] = (pTransaction->nSlaveDeviceAddress<<1) | 0x01;

    // Buffer length; if nBufferLength == 256 we will send 0 as buffer length to firmware
    if ((pTransaction->nBufferLength > 256) || (pTransaction->nBufferLength == 0))
        return U2C_BAD_PARAMETER;
    
    if (pTransaction->nBufferLength == 256)
        req.OutputBuffer[i++] = 0;
    else
        req.OutputBuffer[i++] = (UCHAR)pTransaction->nBufferLength;
    req.OutputBufferLength = i;
    
    req.InputBufferLength = pTransaction->nBufferLength+1;
    rc = u2c_transaction( hDevice, &req );
    if( rc ) return rc;
    
    memcpy(pTransaction->Buffer, req.InputBuffer, pTransaction->nBufferLength);
    return rc;
}    


U2C_RESULT U2C_Write(HANDLE hDevice, PU2C_TRANSACTION pTransaction)
{ 
    U2C_RESULT rc;
    DEVICE_REQUEST req;
    int     j,k,i = 0;
    long    MemoryAddr;

    if (pTransaction->nSlaveDeviceAddress > 0x7F)
        return U2C_BAD_PARAMETER;

    req.OutputBuffer[i++] = 0x02;
    req.OutputBuffer[i++] = pTransaction->nMemoryAddressLength;

    // slave address for write (lsb low)
    req.OutputBuffer[i++] = (pTransaction->nSlaveDeviceAddress<<1) & 0xFE;
    
    // memory address if it exists
    MemoryAddr = pTransaction->nMemoryAddress;
    for (j = i + pTransaction->nMemoryAddressLength - 1; j >= i; j--)
    {
        req.OutputBuffer[j] = (BYTE)MemoryAddr & 0xFF;
        MemoryAddr >>= 8;
    }
    i += pTransaction->nMemoryAddressLength;

    // Buffer length; if nBufferLength == 256 we will send 0 as buffer length to firmware
    if ((pTransaction->nBufferLength > 256) || (pTransaction->nBufferLength == 0))
        return U2C_BAD_PARAMETER;

    if (pTransaction->nBufferLength == 256)
        req.OutputBuffer[i++] = 0;
    else
        req.OutputBuffer[i++] = (UCHAR)pTransaction->nBufferLength;
    
    // Data for output
    for( k = 0; k < pTransaction->nBufferLength; k++)
        req.OutputBuffer[i++] = pTransaction->Buffer[k];
    req.OutputBufferLength = i;
    
    req.InputBufferLength = 1;
    rc = u2c_transaction( hDevice, &req );
    return rc;
}

U2C_RESULT U2C_SetI2cFreq(HANDLE hDevice, BYTE Frequency)
{
    U2C_RESULT rc;
    DEVICE_REQUEST req;

    req.OutputBufferLength=2;
    req.OutputBuffer[0]=0x1B;
    req.OutputBuffer[1]=Frequency;
    req.InputBufferLength=1;

    rc = u2c_transaction( hDevice, &req );
    return rc;
}

U2C_RESULT U2C_GetI2cFreq(HANDLE hDevice, BYTE *pFrequency)
{
    U2C_RESULT rc;
    DEVICE_REQUEST req;

    req.OutputBufferLength=1;
    req.OutputBuffer[0]=0x1C;
    req.InputBufferLength=2;

    rc = u2c_transaction( hDevice, &req );
    *pFrequency = req.InputBuffer[0];
    return rc;
}    

//-----------------------------------------------------------------------------
// Local Functions 
//-----------------------------------------------------------------------------

/*
* Convert errno codes returned by libusb to U2C_RESULT
*/
static U2C_RESULT u2c_libusb_rc( int rc )
{
    if( rc )
        switch( rc )
        {            
        case EBUSY:
            return U2C_DEVICE_BUSY;
        case ENOMEM:
            return U2C_MEMORY_ERROR;
        default:
            return U2C_UNKNOWN_ERROR;
        }
    else
        return U2C_SUCCESS;
}   

/*
* Convert FW status codes to U2C_RESULT
*/
static U2C_RESULT u2c_fw_rc( int rc )
{
    if( rc )
        switch( rc )
        {            
        case 1:
            return U2C_TRANSACTION_FAILED;
        case 2:
            return U2C_SLAVE_OPENNING_FOR_WRITE_FAILED;
        case 3:
            return U2C_SLAVE_OPENNING_FOR_READ_FAILED;
        case 4:
            return U2C_SENDING_MEMORY_ADDRESS_FAILED;
        case 5:
            return U2C_SENDING_DATA_FAILED;
        case 6:
            return U2C_NOT_IMPLEMENTED;
        case 7:
            return U2C_NO_ACK;            
        default:
            return U2C_UNKNOWN_ERROR;
        }
    else
        return U2C_SUCCESS;
}   

//-----------------------------------------------------------------------------
// U2C Transaction Core Function 
//-----------------------------------------------------------------------------
static int u2c_transaction( HANDLE hDevice, PDEVICE_REQUEST req )
{
    int     rc=0;
    sigset_t set, oldset;
    int     i;

/* 
 After interface claimed no other process can access 
 same interface, so this is good for locking     
*/
    
    /* Claim Interface */
    rc = usb_claim_interface( hDevice, DIMAX_IF );
    if( rc )
        return u2c_libusb_rc(rc);
    
/*
 To be sure signal will not kill process between out and in bulk 
 transactions postpone signals. (This will not stop SIGKILL and SIGSTOP)
*/     
    
    /* Block signals to make transaction atomic */
    memset( &set, 0xFF, sizeof(sigset_t) );
    sigprocmask(SIG_BLOCK, &set, &oldset);

    /* Bulk out */
    rc = usb_bulk_write( hDevice, DIMAX_OUT_EP, 
                req->OutputBuffer, req->OutputBufferLength, DIMAX_TIMEOUT );
    if( rc != req->OutputBufferLength )
    {        
        if( debug )
            printf( "usb_bulk_write failed rc=%d\n", rc );
        rc = U2C_TRANSACTION_FAILED;
        goto cleanup;
    }        

    /* Bulk in */
    rc = usb_bulk_read( hDevice, DIMAX_IN_EP, 
                req->InputBuffer, req->InputBufferLength, DIMAX_TIMEOUT );
    if( rc != req->InputBufferLength )
    {        
        if( debug )
            printf( "usb_bulk_read failed rc=%d\n", rc );
        rc = U2C_TRANSACTION_FAILED;
        goto cleanup;
    }        

    /* Check Status. Status is returned by fw in last byte */
    if( req->InputBuffer[req->InputBufferLength-1] )
        rc = u2c_fw_rc( req->InputBuffer[req->InputBufferLength-1] );
    else
        rc = U2C_SUCCESS;

    
cleanup:    
    /* Unblock signals which was not blocked */
    for( i=0; i<sizeof(sigset_t); i++ )
        ((char*)(&set))[i]=~((char*)(&oldset))[i];
    sigprocmask(SIG_UNBLOCK, &set, &oldset);

    /* Release Interface */
    usb_release_interface( hDevice, DIMAX_IF ); 
    return rc;
}

